#=========================================================================
#						==== ArsBaseClass ====
#
# This is the basis of an object-oriented approach to ARSPerl.  My goal
# was to simpify access to ARS schemata (see SYNOPSIS below).  This isn't
# necessarily an efficient interface to ARS.
#
# This is certainly not a complete interface to all ARS field types
# and functionality, but it shouldn't be hard to extend it...
#
# Currently, ArsBaseClass only has the ability to perform trivial queries
# (i.e., on a single field) because this was all I needed.  If this is a
# problem for you, you will probably want to implement functionality for
# more complex queries.  If you come up with something elegant, please
# tell me about it!
#
# If you're concerned about speed, this probably isn't what you want.
#
# ArsBaseClass must be used by a derived class. 
# 
# The philosophy is as follows:
#
# one runs the script "ars-gen-subclass" on a particular schema.  The
# script generates the sub-class and writes it to STDOUT.  Before you
# run "ars-gen-subclass", you should decide whether you want the
# get/set methods (which correspond to the field names) to have the
# format
#
#	this_is_a_get_set_method
#
# or
#
#	thisIsAGetSetMethod
#
# and set $METHOD_STYLE in ars-gen-subclass accordingly.
#
# Also before you run "ars-gen-subclass" edit this file and set the
# $ARSUSER, $ARSPASS, and $ARSSERVER variables appropriately.
#
# you need to edit the generated sub-class file and check (at least) four
# things:
#
# 1. ars-gen-subclass creates private labels for the schema fields.  These
#    should be unique and reasonable.  Change them to whatever you like.
#    It is these that will be used as the names of access methods.  Keep
#    in mind that if you change these labels, you will have to change
#    then again if you want to re-run ars-gen-subclass.
#
# 2. set "query => 1" for the fields that you want to be queried when
#    the getRecord() method is called.  All other fields will only be
#    queried when the corresponding access method is called (delayed
#    initialization).
#
# 3. "query_field" should be set to the private label of the field
#    which is to be used for selection.  In general this should be a
#    field whose value must be unique -- thus returning a single
#    record.  If there is no such field, use the record-ID, which
#    then makes getRecord() equivalent to getRecordById(), if
#    somewhat less efficient.
#
# 4. You'll probably want to change the package name.
#
# SYNOPSIS:
#
#	use DerivedClass;
#
#	$value = 'something';
#
#	$record = DerivedClass->new;
#	$record->getRecord($value) || die "query failed for val=$value";
#
#	# change values
#	$record->customer_name('Mr. Bean');
#	$record->favorite_game('bingo');
#
#	# write back to ARS
#	$record->write;
# 
#   If you have set "query_field" to something reasonable, you can
#   condense the calls to the constructor and to getRecord() as follows:
#
#   $record = DerivedClass->new($value) || die "query failed for $value";
#
# If you perform a query and then change values, an ARS update will
# be performed when the write() method is called.  If you get a new
# object, and then set values using access methods without performing
# a query beforehand, when the write() method is called a new ARS
# record will be created.  In this case, the record-ID is returned
# by the write() method.
#
# Keep in mind that setting "query => 0" in CLASS_DATA for a subclass
# does *not* mean that that field value will never be available to you.
# It simply means a standard record-fetch will not retrieve that field
# value immediately.  If you later call the get-method for such a field
# an automatic fetch will be performed for you transparently and you
# will magically be supplied the value.  By default "ars-gen-subclass"
# marks all "diary" fields for delayed initialization.
#
# Because a schema can change after ars-gen-subclass has been used
# to dump its structure, ArsBaseClass implements the method
# "schemaChanged" to compare the dumped structure in a derived
# class to the structure in the ARS schema.  schemaChanged()
# returns nothing if the structures are consistent, and a
# scalar consisting of descriptions of inconsistencies otherwise.
#
# ### GENERIC QUERY ###
#
# A list of object can be obtained using the mechanism:
#
#	@list = subclass::query(ars_field1 => val1 [, ars_field2 => val2, ...]);
#
# This causes a trivial ARS query to be performed where:
# 1. each field in a query-pair must be equal to the value
# 2. the pairs are ANDed together.
# 
# ### DEFAULT USER ###
#
# By default, ArsBaseClass performs an ARS login and maintains
# a control object for use by subclasses.  You can override the default
# (default) user in ArsBaseClass by doing:
#
#	use ArsBaseClass USER => 'myuser', PASS => 'mypass';
#
# You can also use the same mechanism for a subclass to access the
# corresponding schema as a different user.
#
# ### PECULIARITIES ###
#
# There are some peculiarities of accessing ARS objects which
# bear description.
#
# - If you call write() but have not modified any fields, write() does
#	nothing.
#
# - the write() method *only* actually writes *changed* data to ARS.
#
# - If you create an object, modify fields, and call write():
#
#		$obj = SubClass->new($query_value);
#		$obj->some_field("new-value");
#		$obj->write;
#
#	The data is written and the object is "reset", which means the
#	retrieved ARS data is thrown away.  However, the object "knows"
#	its record_id, such that if you continue to use it, it will behave
#	as you'd expect:
#
#	- if you read from a field, a transparent (new) query to ARS will be
#	  performed.
#
#	- of course you can modify fields and write() your changes
#
# - fields of type "diary" present a bit of a problem.  The field-value
#	returned by a get-method on an object simply returns what ars_GetEntry()
#	returns, a reference to an array containing references to hashes, which,
#	in turn, contain the keys "user", "timestamp", and "value". So
#	what happens when you write a new entry to a diary field and then
#	immediately retrieve that field, i.e.:
#
#		$obj = SubClass->new($query_value);
#		$obj->diary_field("this is a new entry");
#		$foo = $obj->diary_field;
#
#	In this case I "fake up" an array which includes the new entry.  The
#	"user" and "timestamp" values are 'n/a' (not available).  Presumably
#	these could be set to the user used to log in to ARS and the current
#	epoch-seconds value.
#
# - Attachments
#
#   You can use the standard get/set methods for attachments:
#
#   - get
#
#     $attachment = $object->attachment_field1;
#
#     $attachment is a reference to a hash containing two keys:
#       name   => 'the name of the attachment'
#       value  => <reference to a scalar containing text>
#
#   - set
#
#     $ref = {
#       type  => 'buffer',
#       name  => 'some arbitrary string',
#       value => reference to scalar containing text
#     };
#     $object->attachment_field1($ref);
#
#
#		**************************************
#		*************** Methods **************
#		**************************************
#
#	write()
#		writes *modified data only* to ARS.  If called after a query
#		(i.e., after object has been loaded from ARS) then an update
#		is performed.  Otherwise a create is performed.  Returns
#		nothing after an update, the record-ID after a create.
#
#	getRecord()
#		loads data into object from ARS from the record retrieved
#		using "query_field" in CLASS_DATA.  Returns 1 on success,
#		undef on failure.
#
#	getRecordById()
#		loads data into object from ARS using the record-ID.  Returns 1
#		on success, undef on failure.
#
#	query()
#		This should generally be called from a derived class.  parameters
#		are a field-name and value.  Returns the record-ID of the first
#		match in scalar context, or a list of record-IDs in list context.
#		On failure returns undef in scalar context and empty list in list
#		context.
#
#	reload()
#		Causes all data stored in object to be thrown away and ARS data
#		to be reloaded into object.
#
#	recordModified()
#		fetches the "modified time" field from ARS and compares it to the
#		modifed-time field stored in the object.  Returns 1 if the ARS
#		data is newer than the object data.
#
#	setFields()
#		This does an end-run around the normal get/set methodology.
#		The intent was to allow a large number of fields to be
#		modified at one time without having to call the individual
#		get/set "methods".  The input is a reference to a hash whose
#		keys are valid field-names and whose corresponding values
#		are the new values for those fields.  Returns nothing.
#
#	getFields()
#		even uglier than setFields().  Retrieves all loaded or modified
#		fields and returns a reference to a hash as described above for
#		setFields().
#
#	schemaChanged()
#		checks whether the CLASS_DATA structure still corresponds to
#		the ARS schema.  If there are discrepencies, a scalar is returned
#		containing a description of all discrepencies.  Otherwise, returns
#		empty scalar.
#
#	recordId()
#		returns the ARS request-ID if known, i.e., after a write() or
#		after a query. Otherwise returns undef.
#
#
# you will probably want to replace logmsg() with something more
# appropriate.
#
# ** if you find this useful, please let me know! **
#
# Rob Urban <urban@tru64.org>
# May, 2003
#=========================================================================

package ArsBaseClass;

use Carp;
use ARS;
use File::Basename;
use strict;

#---------------------------------------------------------------
# these are private unrelated modules.  replace or delete.
#---------------------------------------------------------------
use Debug;

my @CTL_LIST	= ();
my $CTL;		# used to store the control structure of an open connection

#my $DEBUG = 0;
my $VERBOSE = 0;
my $ARSDEBUG = 0;

my $DEBUG = Debug::level();
my $ARS_WRITE = Debug::arsWrite();
#my $VERBOSE = Debug::verbosity();

#---------------------------------------------------------------
# global default values for USER, PASS, SERVER
# set $ARSUSER to undef() to avoid automatic login
#---------------------------------------------------------------
my $ARSUSER		= 'your-ars-user-here';
my $ARSPASS		= 'your-ars-password-here';
my $ARSSERVER	= 'your-server-name-here';
my $MAXRETRIEVE	= 0;
my $PRIMARY_KEY	= 1;
my $MODIFY_TIME	= 6;
my $VERSION		= 1.1;
my %LOGIN_DB;

$VERBOSE && print "ArsBaseClass: using SERVER=$ARSSERVER\n";

sub import
{
	shift;	# get rid of $class
	#foreach my $word (@_) {
	my $word;
	$ARSDEBUG && print "ArsBaseClass::import()\n";
	while ($word = shift) {
		if ($word eq 'USER') {
			$ARSUSER = shift;
			$ARSDEBUG && print "setting ARSUSER=[$ARSUSER]\n";
		} elsif ($word eq 'PASS') {
			$ARSPASS = shift;
			$ARSDEBUG && print "setting ARSPASS=[$ARSPASS]\n";
		} elsif ($word eq 'SERVER') {
			$ARSSERVER = shift;
			$ARSDEBUG && print "setting ARSSERVER=[$ARSSERVER]\n";
		} else {
			die "unknown directive [$word]";
		}
	}
}

#---------------------------------------------------------------
# login
#
# This is not a subroutine because I want _one_ login when this
# module is loaded
#---------------------------------------------------------------
INIT {
	if (defined($ARSUSER)) {
		$VERBOSE && logmsg("logging in to ARS (user=$ARSUSER)...");
		$ARSDEBUG && print "logging in to ARS as $ARSUSER,$ARSPASS,$ARSSERVER...\n";
		#print("logging in to ARS (user=$ARSUSER)...\n");
		eval {
			($CTL = ars_Login($ARSSERVER, $ARSUSER, $ARSPASS))
				|| die "login failed";
				$LOGIN_DB{$CTL} = {
					user	=> $ARSUSER,
					pass	=> $ARSPASS,
					server	=> $ARSSERVER,
				};
		};
		if ($@) {
			die "ARS login failed. $ARSUSER:$ARSPASS\@$ARSSERVER\nerror=$ars_errstr";
		}
		push(@CTL_LIST, $CTL);
	}
}

#---------------------------------------------------------------
# ... and log out when done.
#---------------------------------------------------------------
END {
	foreach my $ctl (@CTL_LIST) {
		if ($VERBOSE) {
			my $user = $LOGIN_DB{$ctl}->{user};
			logmsg("logging out of ARS... [user=$user]");
		}
		ars_Logoff($ctl);
	}
	#print STDERR "done.\n";
}

sub new
{
	my $proto = shift;
	my $class = ref($proto) || $proto;

	my $self = {
		loaded	=> 0,
		dirty	=> 0,
		ctl		=> $CTL,
	};

	bless($self, $class);

	my $ctl;

	my @args;
	while (@_) {
		$_ = shift;
		if (/^CTL/) {
			$ctl = shift;
			$self->{ctl} = $ctl;
		} else {
			push(@args, $_);
		}
	}

	($VERBOSE > 1) && logmsg("  using CTL=[$self->{ctl}] for this object");

	if (@args) {
		$VERBOSE && logmsg("  ArsBaseClass::new(): calling getrecord on [$args[0]]");
		#if (!defined($self->getRecord(shift(@args)))) {
		if (!defined($self->getRecord(@args))) {
			$VERBOSE && logmsg("  query failed. returning undef.");
			return undef;
		}
	}

	$self;
}

sub login
{
	my ($server, $user, $password);
	#print "login(): params=(",join(',', @_),"\n";
	while (@_) {
		$_ = shift;
		if (/^SERVER/) {
			$server = shift;
		} elsif (/^USER/) {
			$user = shift;
		} elsif (/^PASS/) {
			$password = shift;
		} else {
			die "bad param: [$_]";
		}
	}

	if (!$user || !$password) {
		die "user/password missing";
	}

	if (!$server) {
		#print "using default SERVER=$ARSSERVER\n";
		$server = $ARSSERVER;
	}

	my $ctl;
	$VERBOSE && logmsg("logging in to ARS (user=$user)...");

	#print("logging in to ARS (user=$user)...\n");
	#stackTrace();

	($ctl = ars_Login($server, $user, $password))
		|| die "ARS login for [$user] failed";

	$LOGIN_DB{$ctl} = {
		user	=> $user,
		pass	=> $password,
		server	=> $server,
	};
	push(@CTL_LIST, $ctl);
	return $ctl;
}

#------------------------------------------------------------------
# _getClassData is a placeholder. The real _getClassData() must be
# implemented in the subclass
#------------------------------------------------------------------
sub _getClassData
{
	die "_getClassData() not implemented";
}

#------------------------------------------------------------------
# This is called by the subclasses to generate reverse mappings
# for the private hash keys and also to generate a list of fields
# that should be retrieved by getRecord()
#
# It expects to find a hashref at $cd->{priv_keys}
#
# It will write to $cd->{reverse_map} and $cd->{query_list}
# and also to $cd->{priv_keys}->{$key}->{enum_rev} if the
# field is of type 'enum'
#------------------------------------------------------------------
sub init
{
	my $cd = shift;

	$VERBOSE && logmsg("** running init() for schema = [$cd->{schema}] **");

	# generate field list
	my @l;
	my $rev;
	foreach my $key (keys(%{$cd->{priv_keys}})) {
		#print "KEY [$key]\n";
		if (($cd->{priv_keys}->{$key}->{query}) ||
			($cd->{priv_keys}->{$key}->{id} eq $MODIFY_TIME))
		{
			push(@l, $cd->{priv_keys}->{$key}->{id});
		}
		#my $rk = $cd->{priv_keys}->{$key}->{id};
		#print "setting rev->{$rk} = [$key]\n";;
		$rev->{$cd->{priv_keys}->{$key}->{id}} = $key;

		# and the reverse-maps for the enums
		if ($cd->{priv_keys}->{$key}->{type} eq 'enum') {
			my $r;
			foreach my $ekey (keys(%{$cd->{priv_keys}->{$key}->{enum}})) {
				$r->{$cd->{priv_keys}->{$key}->{enum}->{$ekey}->[0]} = $ekey;
			}
			$cd->{priv_keys}->{$key}->{enum_rev} = $r;
		}
	}
	$cd->{reverse_map} = $rev;
	$cd->{query_list} = \@l;

	#dumpClassdata($cd, "end of init()");
}

#------------------------------------------------------------------
# AUTOLOAD() is used to get or set object field-values.  This is
# done by referring to the hash-keys in $CLASS_DATA->{priv_keys}
# as methods.  Since there are no methods by those names defined,
# AUTOLOAD() is called and the name of the called method is put
# in $AUTOLOAD.  AUTOLOAD() then checks if there is in fact a key
# in the {priv_keys} hash with that name.  If not, it aborts with
# an error message.
#------------------------------------------------------------------
sub AUTOLOAD
{
	my $self = shift;
	no strict 'vars';

	my $field = $AUTOLOAD;
	$field =~ s/^.*::([^:]+)$/$1/;		# strip package name
	return if ($field eq 'DESTROY');

	if (!ref($self) || ($self !~ /=/)) {
		confess "[private] Undefined subroutine $AUTOLOAD called at";
	}

	my $cd = $self->_getClassData;

	if (!exists($cd->{priv_keys}->{$field})) {
		confess "[private] Undefined subroutine $AUTOLOAD called at";
	}

	if (@_) {
		# setting value
		my $val = shift;
		$self->{changed_data}->{$field} = $val;
		if ($VERBOSE) {
			if ($val =~ /\n/) {
				# for debugging, take only first 4 lines
				my @f = split(/\n/, $val);
				my $l = (@f < 4) ? $#f : 3;
				$val = "\n\t".join("\n\t", @f[0..$l])."\n  ";
			}
			logmsg("  SET-FIELD: [$field] = [$val]");
		}
		$self->{dirty} = 1;
		return;
	} else {
		return $self->getField($field);
	}
}

#------------------------------------------------------------------
# getField() is mostly for the use of AUTOLOAD(), but you may also
# use it :).
#
# It does many things:
#
# - it calls die() if an attempt is being made to read a field-value
#	before any data have been loaded from ARS (and the field has not
#	been modified)
#
# - it checks if an intervening write() was performed which would
#	require a re-loading of ARS data
#
# - it checks if a field-value has been requested that is marked
#	for delayed-initialization, and if so, performs a query to ARS
#	to get the data
#
# - it checks if a diary-field is being requested; this requires
#	special handling
#
# - finally, it checks if the field-value has been modified, and if
#	so, returns the modified value, otherwise it returns the
#	ARS-queried value
#------------------------------------------------------------------
sub getField
{
	my ($self, $field) = @_;

	my $cd = $self->_getClassData;

	# check if trying to read from an uninitialized object
	if (!exists($self->{changed_data}->{$field}) && !exists($self->{record_id}))
	{
		die "trying to read [$field] from uninitialized object";
		#return undef;
	}

	# check if data must be reloaded because of an intervening write()
	if (($self->{loaded} == 0) && (exists($self->{record_id}))) {
		$self->getRecordById($self->{record_id});
	}

	# getting value
	if ($self->{loaded} && !exists($self->{ars_data}->{$field})) {
		# this field had deferred initialization set
		$self->_queryField($field);
	}

	#----------------------------------------------------------
	# special case for 'diary' fields
	#----------------------------------------------------------
	if ($cd->{priv_keys}->{$field}->{type} eq 'diary') {
		my $aref;
		if (ref($self->{ars_data}->{$field})) {
			push(@{$aref}, @{$self->{ars_data}->{$field}});
		}
		if (exists($self->{changed_data}->{$field})) {
			push(@{$aref}, {
				user		=> 'n/a',
				timestamp	=> 'n/a',
				value		=> $self->{changed_data}->{$field},
			});
		}
		return $aref;
	}

	#----------------------------------------------------------
	# special case for 'attach' fields (grrrrrrrrrr)
	#----------------------------------------------------------
	if ($cd->{priv_keys}->{$field}->{type} eq 'attach') {
		if (exists($self->{changed_data}->{$field})) {
			return {
				name	=> '',
				value	=> $self->{changed_data}->{$field},
			};
		} else {
			return $self->{ars_data}->{$field};
		}
	}

	#----------------------------------------------------------
	# if field has been modified take modified value, otherwise
	# take value from ARS
	#----------------------------------------------------------
	if (exists($self->{changed_data}->{$field})) {
		# setting value
		return $self->{changed_data}->{$field};
	} elsif (exists($self->{ars_data}->{$field})) {
		return $self->{ars_data}->{$field};
	}

	# catch-all
	return undef;
}

#------------------------------------------------------------------
# _queryField() handles delayed-initialization.
#------------------------------------------------------------------
sub _queryField
{
	my ($self, $field) = @_;

	$ARSDEBUG && print "delayed loading of field [$field] ...\n";

	my $cd = $self->_getClassData;
	my $schema = $cd->{schema};

	my $fid = $cd->{priv_keys}->{$field}->{id};

	my $rid = $self->{record_id};

	my ($key, $value) = ars_GetEntry($self->{ctl}, $schema, $rid, $fid);

	# a little sanity-check
	if ($key ne $fid) { confess "whoa! schema=[$schema], field=[$field], key=[$key], fid=[$fid]"; }

	# if there is a key in %h == undef, error
	if (!defined($key)) {
		return undef;
	}

	if ($cd->{priv_keys}->{$field}->{type} eq 'enum') {
		$self->{ars_data}->{$field} = 
			$cd->{priv_keys}->{$field}->{enum_rev}->{$value};
	} elsif ($cd->{priv_keys}->{$field}->{type} eq 'attach') {
		if (!ref($value)) {
			$self->{ars_data}->{$field} = undef;
			return;
		}
		my $a = ars_GetEntryBLOB($self->{ctl}, $schema, $rid, $key, ARS::AR_LOC_BUFFER);
		$self->{ars_data}->{$field} = {
			name	=> $value->{name},
			value	=> \$a,
		};
	} else {
		$self->{ars_data}->{$field} = $value;
	}
}

sub getRecord
{
	my $self = shift;

	my $cd = $self->_getClassData;

	my $qf = $cd->{query_field};
	my $qf2 = $cd->{query_field2};
	my $schema = $cd->{schema};

	my @params = ($qf => shift);
	if (@_) {
		defined($qf2) || die "getRecord called with 2 params but qf2 not defined";
		push(@params, $qf2 => shift);
	}

	$VERBOSE && logmsg("  getRecord: query for @params");

	my $rid = query($cd, $self->{ctl}, @params);

	$VERBOSE && logmsg("  getRecord: query return rid=[$rid]");

	if (!defined($rid)) {
		return undef;
	}

	if (!defined($self->getRecordById($rid))) {
		die "getRecord: record disappeared";
	}

	return 1;
}

sub getAllRecordIds
{
	my $ctl = (ref($_[0]) eq 'ARControlStructPtr') ? shift : $CTL;
	my $cd = shift;

	my $qf = $cd->{query_field};
	my $schema = $cd->{schema};

	my @rids = query($cd, $ctl, 'ALL');

	return @rids;
}

#------------------------------------------------------------------
# getRecordById() is THE function that is responsible for loading
# ARS data into an object.
#------------------------------------------------------------------
sub getRecordById
{
	my ($self, $rid) = @_;

	# reset record
	$self->{ars_data}		= {},	# ref to hash of ARS field data
	$self->{changed_data}	= {},	# ref to hash of fields updated
	$self->{dirty}			= 0,	# 1 if data needs to be written
	$self->{loaded}			= 0,

	my $data = $self->_getArsData($rid);
	if (!defined($data)) {
		delete($self->{record_id});
		return undef;
	}

	$self->{ars_data} = $data;
	$self->{loaded} = 1;
	$self->{record_id} = $rid;

	return 1;
}

#------------------------------------------------------------------
# query() fires a query at ARS and returns a hashref of key-value
# pairs.  If called in a scalar context, returns the first record-ID.
#
# If called in a list context, returns a list of record-IDs.
#------------------------------------------------------------------
sub query
{
	#my ($cd, $key, $value) = @_;		# $cd contains a ref to CLASS_DATA
	my $cd = shift;
	my $ctl = (ref($_[0]) eq 'ARControlStructPtr') ? shift : $CTL;
	
	#print "cd=[$cd]\n";
	my $schema = $cd->{schema};

	my $qualifier;

	my ($key, $value);

	if ($_[0] eq 'ALL') {
		$qualifier = qq[(1 = 1)];
	} else {
		my (@qlist);
		while(@_) {
			($key, $value) = (shift, shift);

			if (!exists($cd->{priv_keys}->{$key})) {
				#die "query: key [$key] unknown!";
				confess "query: key [$key] unknown!";
			}

			my $field_id = $cd->{priv_keys}->{$key}->{id};

			if (!defined($value)) {
				#$qualifier = "'$field_id' = NULL";
				push(@qlist, qq[('$field_id' = NULL)]);
			} elsif ($cd->{priv_keys}->{$key}->{type} eq 'enum') {
				#if (defined($value) && exists($cd->{priv_keys}->{$key}
				$value = _lookupEnum($cd, $key, $value);
				#$qualifier = qq['$field_id' = $value];
				#push(@qlist, qq[('$field_id' = "$value")]);
				push(@qlist, qq[('$field_id' = $value)]);
			} else {
				#$qualifier = qq['$field_id' = "$value"];
				push(@qlist, qq[('$field_id' = "$value")]);
			}

			$qualifier = join(' AND ', @qlist);
		}
	}

	my $q;
	# print "Q=[$qualifier]\n";
	#($q = ars_LoadQualifier($CTL, $schema, $qualifier)) ||
	#    die $ars_errstr;

	$q = ars_LoadQualifier($ctl, $schema, $qualifier);
	if (!$q) {
	    confess "error in ars_LoadQualifier.\nqualifier=[$qualifier]\nars_errstr=[$ars_errstr]";
	}

	my @list = ars_GetListEntry($ctl, $schema, $q, $MAXRETRIEVE, 0);
	if (!wantarray) {
		if (!@list) {
			return undef;
		} else {
			return $list[0];
		}
	}

	# if @list only has one element, or if first element is undef, error
	#if (($#list == 0) || !defined($list[0])) { return (); }
	my ($rid, $desc, @out);
	while (@list) {
		($rid, $desc, @list) = @list;
		push(@out, $rid);
	}

	return (@out);
}

sub genericQuery
{
	print "huhu\n";
}

sub old_query
{
	my ($cd, $key, $value) = @_;		# $cd contains a ref to CLASS_DATA
	
	my $schema = $cd->{schema};

	my $qualifier;
	my $field_id = $cd->{priv_keys}->{$key}->{id};
	if ($key eq 'ALL') {
		$qualifier = qq[(1 = 1)];
	} elsif (!defined($value)) {
		$qualifier = "'$field_id' = NULL";
	} elsif ($cd->{priv_keys}->{$key}->{type} eq 'enum') {
		#if (defined($value) && exists($cd->{priv_keys}->{$key}
		$value = _lookupEnum($cd, $key, $value);
		$qualifier = qq['$field_id' = $value];
	} else {
		$qualifier = qq['$field_id' = "$value"];
	}

	my $q;
	# print "Q=[$qualifier]\n";
	($q = ars_LoadQualifier($CTL, $schema, $qualifier)) ||
	    die $ars_errstr;

	my @list = ars_GetListEntry($CTL, $schema, $q, $MAXRETRIEVE, 0);
	if (!wantarray) {
		if (!@list) {
			return undef;
		} else {
			return $list[0];
		}
	}

	# if @list only has one element, or if first element is undef, error
	#if (($#list == 0) || !defined($list[0])) { return (); }
	my ($rid, $desc, @out);
	while (@list) {
		($rid, $desc, @list) = @list;
		push(@out, $rid);
	}

	return (@out);
}

sub ctl
{
	my $self = shift;

	return $self->{ctl};
}

sub _getArsData
{
	my ($self, $rid) = @_;

	my $res;

	my $cd = $self->_getClassData;
	my $schema = $cd->{schema};
	my $ql = $cd->{query_list};

	$VERBOSE && logmsg("  query: schema = [$schema], rid=[$rid]");
	my %h = ars_GetEntry($self->{ctl}, $schema, $rid, @{$ql});
	
	# if there is a key in %h == undef, error
	if (exists($h{undef()})) {
		return undef;
	}

	# create new hash using private field names
	foreach my $key (keys(%h)) {
		my $pkey = $cd->{reverse_map}->{$key};
		if ($cd->{priv_keys}->{$pkey}->{type} eq 'enum') {
			# this should work with ARS $NULL$ values, as it simply
			# propagates the undef
			my $tmp = $cd->{priv_keys}->{$pkey}->{enum_rev}->{$h{$key}};
			$res->{$pkey} = $tmp;
		} elsif ($cd->{priv_keys}->{$pkey}->{type} eq 'attach') {
			# do not call ars_GetEntryBLOB() if there is no attachment here
			if (!ref($h{$key})) { next; }
			my $a = ars_GetEntryBLOB($self->{ctl}, $schema, $rid, $key,
				ARS::AR_LOC_BUFFER);
			$res->{$pkey} = {
				name => $h{$key}->{name}, 
				value	=> \$a,
			};
		} else {
			$res->{$pkey} = $h{$key};
		}
	}

	return $res;
}

sub dumpData
{
	my $self = shift;

	print "-- DUMP DATA --\n";
	if (exists($self->{ars_data})) {
		print "  ARS_DATA:\n";
		foreach my $key (keys(%{$self->{ars_data}})) {
			if (defined($self->{ars_data}->{$key})) {
				printf("\t%-30s %s\n", $key, $self->{ars_data}->{$key});
			} else {
				printf("\t%-30s NULL\n", $key);
			}
		}
	}
	if (exists($self->{changed_data})) {
		print "  CHANGED_DATA:\n";
		foreach my $key (keys(%{$self->{changed_data}})) {
			if (defined($self->{changed_data}->{$key})) {
				printf("\t%-30s %s\n", $key, $self->{changed_data}->{$key});
			} else {
				printf("\t%-30s NULL\n", $key);
			}
		}
	}
}

sub dumpHash
{
	my ($hr, $label) = @_;

	print "-- DUMP HASH [$label] --\n";

	foreach my $key (keys(%{$hr})) {
		printf("    %12s  %s\n", $key, $hr->{$key});
	}
}

#------------------------------------------------------------------
# _convertToArsFormat() is passed a reference to a hash with keys
# matching those in FIELDS.  The keys are converted to Field-IDs, and
# in the case of enum fields, the values are converted to the numbered
# equivalents
#------------------------------------------------------------------
sub _convertToArsFormat
{
	my ($cd, $href) = @_;		# $cd contains a ref to CLASS_DATA

	my $ref;

	foreach my $key (keys(%{$href})) {
		if (!exists($cd->{priv_keys}->{$key})) {
			die "_convertToArsFormat: [$key] not found";
		}
		my $type = $cd->{priv_keys}->{$key}->{type};
		my $fid = $cd->{priv_keys}->{$key}->{id};
		#$ARSDEBUG && print " ## converting [$key] --> [$fid]\n";
		if ($type eq 'enum') {
			if (defined($href->{$key}) &&
				(!exists($cd->{priv_keys}->{$key}->{enum}->{$href->{$key}})))
			{
				confess "no such value [$href->{$key}] in enum field [$key]";
			}
			$ref->{$fid} = $cd->{priv_keys}->{$key}->{enum}->{$href->{$key}}->[0];
		} elsif ($type eq 'attach') {
			if (ref($href->{$key}) ne 'HASH') {
				die "value for attachment field not hashref";
			}
			if (exists($href->{$key}->{name})) {
				$ref->{$fid}->{name} = $href->{$key}->{name};
			}
			exists($href->{$key}->{type}) ||
				die "type key missing from attachment hash";

			if ($href->{$key}->{type} eq 'file') {
				my $file = $href->{$key}->{file};
				(-R $file) || die "file [$file] doesn't exist or unreadable";
				$ref->{$fid}->{size} = -s $file;
				$ref->{$fid}->{file} = $file;
			} else {
				$ref->{$fid}->{size} = length(${$href->{$key}->{value}});
				$ref->{$fid}->{buffer} = ${$href->{$key}->{value}};
			}
		} else {
			$ref->{$fid} = $href->{$key};
		}
	}

	$ref;
}

sub _lookupEnum
{
	my ($cd, $field, $value) = @_;

	if (defined($value) &&
		!exists($cd->{priv_keys}->{$field}->{enum}->{$value}))
	{
		confess "no such value [$value] in enum field [$field]";
	}
	return $cd->{priv_keys}->{$field}->{enum}->{$value}->[0];
}

sub delete
{
	my $self = shift;

	my $cd = $self->_getClassData;
	my $schema = $cd->{schema};

	exists($self->{record_id}) || die "cannot delete before query";

	if (!$ARS_WRITE) {
		logmsg("** aborting ARS delete because ARS_WRITE not set **");
		return;
	}

	ars_DeleteEntry($self->{ctl}, $schema, $self->{record_id})
		|| die "delete: $ars_errstr";
}

sub write
{
	my $self = shift;

	my $cd = $self->_getClassData;
	my $schema = $cd->{schema};

	my $pk_lab = $cd->{reverse_map}->{$PRIMARY_KEY};

	if (exists($self->{changed_data}->{$pk_lab})) {
		die "attempt to set/change primary key during update/create";
	}

	my $converted = _convertToArsFormat($cd, $self->{changed_data});

	($ARSDEBUG > 1) && dumpHash($converted, 'converted');

	if (!$ARS_WRITE) {
		logmsg("** aborting ARS modify because ARS_WRITE not set **");
		return;
	}

	my $eid;

	if (exists($self->{record_id})) {
		# this is an UPDATE
		#my $pkey = $self->{ars_data}->{$pk_lab};
		if (!$self->{dirty}) {
			# no fields updated. do nothing.
			$VERBOSE && logmsg("  write: record not dirty. no work to do.");
			return;
		}
		my $pkey = $self->{record_id};
		($ARSDEBUG > 1) && dd($converted);
		my $ret;
		$ret = ars_SetEntry($self->{ctl}, $schema, $pkey, 0, %{$converted});

		if (!$ret) {
			die "ars_SetEntry: $ars_errstr";
		}
		return;
	} else {
		# this is a CREATE
		($eid = ars_CreateEntry($self->{ctl}, $schema, %{$converted})) || 
			confess "schema=$schema\n$ars_errstr";

		#--------------------------------------------------------
		# set the "record_id" field so the next write() causes an
		# update to be performed
		#--------------------------------------------------------
		$self->{record_id} = $eid;
	}

	
	# reset 
	$self->{ars_data}		= {},	# ref to hash of ARS field data
	$self->{changed_data}	= {},	# ref to hash of fields updated
	$self->{dirty}			= 0,	# 1 if data needs to be written
	$self->{loaded}			= 0,

	# return the request_id
	$eid;
}

sub recordId
{
	my $self = shift;

	return exists($self->{record_id}) ? $self->{record_id} : undef;
}

sub newValue
{
	my ($self, $field) = @_;

	#my $cd = $self->_getClassData;

	if (exists($self->{changed_data}->{$field})) {
		return 1;
	}

	return 0;
}

sub reload
{
	my $self = shift;

	exists($self->{record_id})
		|| die "cannot reload an object that was never loaded";

	return $self->getRecordById($self->{record_id});
}

sub dd
{
	my $href = shift;

	print "** DUMP OF converted changed_data **\n";
	foreach my $key (keys(%{$href})) {
		print "\t$key = [$href->{$key}]\n";
	}
}

#------------------------------------------------------------------
# this method returns true if a record has been updated since the
# time it was read by this process.  Don't need it just now...
#------------------------------------------------------------------
sub recordModified
{
	my $self = shift;

	my $cd = $self->_getClassData;
	my $schema = $cd->{schema};

	my $pk_lab = $cd->{reverse_map}->{$PRIMARY_KEY};
	my $pkey = $self->{ars_data}->{$pk_lab};

	$ARSDEBUG && logmsg("checking if entry [$pkey] changed...");

	my $mt_lab = $cd->{reverse_map}->{$MODIFY_TIME};
	my $mt = $self->{ars_data}->{$mt_lab};

	# get modification time for entry
	my ($lab, $val);
	(($lab, $val) = ars_GetEntry($self->{ctl}, $schema, $pkey, $MODIFY_TIME))
		|| die $ars_errstr;
	
	$ARSDEBUG && logmsg("old mt = [$mt], new mt = [$val]");

	if ($mt eq $val) {
		# no change
		return 0;
	}

	return 1;
}

#------------------------------------------------------------------
# setField()
#------------------------------------------------------------------
sub setField
{
	my ($self, $field, $value) = @_;

	my $cd = $self->_getClassData;

	$VERBOSE && logmsg("  SET-FIELD: [$field] = [$value]");

	if (!exists($cd->{priv_keys}->{$field})) {
		confess "field doesn't exist [$field]";
	}

	$self->{changed_data}->{$field} = $value;
	$self->{dirty} = 1;
}

#------------------------------------------------------------------
# setFields() can be used to set a number of fields using
# a values obtained from a hash.  It sort-of does an end-run
# around the usual get/set methods.
#------------------------------------------------------------------

sub setFields
{
	my ($self, $href) = @_;

	my $cd = $self->_getClassData;

	$VERBOSE && logmsg("setFields:");

	foreach my $key (keys(%{$href})) {
		if (!exists($cd->{priv_keys}->{$key})) {
			confess "field doesn't exist [$key]";
		}
		$VERBOSE && logmsg("\t$key = $href->{$key}");
		$self->{changed_data}->{$key} = $href->{$key};
	}

	$self->{dirty} = 1;
}


#------------------------------------------------------------------
# getFields() really shouldn't be used...
#
# returns a reference to a hash of all loaded or modified fields.
# Modified fields take precedence over values loaded from ARS. The
# value of a diary field will be either a reference to an array of
# hashes if the data comes from ARS, or a simple scalar if the field
# has been modified.
#------------------------------------------------------------------
sub getFields
{
	my $self = shift;

	my (%h1, %h2);

	if (exists($self->{ars_data})) {
		%h1 = %{$self->{ars_data}};
	}

	if (exists($self->{changed_data})) {
		%h2 = %{$self->{changed_data}};
	}

	#return (keys(%{{%h1, %h2}}));
	return {%h1, %h2};
}

#------------------------------------------------------------------
# dump() will dump to STDOUT the contents of an object
#------------------------------------------------------------------
sub dump
{
	my ($self, $all) = @_;

	my $cd = $self->_getClassData;

	my $byid = sub {
		return $cd->{priv_keys}->{$a} <=> $cd->{priv_keys}->{$b};
	};

	foreach my $field (sort $byid keys(%{$cd->{priv_keys}})) {
		my $type = $cd->{priv_keys}->{$field}->{type};
		if ($type eq 'diary') {
			# don't dump diary-fields
			next;
		}
		if (exists($self->{changed_data}->{$field})) {
			printf("  %-25s = %s\n", $field,
				dumpFormat($cd, $field, $self->{changed_data}->{$field}));
		} elsif (exists($self->{ars_data}->{$field})) {
			printf("  %-25s = %s\n", $field,
				dumpFormat($cd, $field, $self->{ars_data}->{$field}));
		} elsif ($all) {
			if ($type eq 'time') {
				printf("  %-25s = \n", niceDate($field));
			} else {
				printf("  %-25s = \n", $field);
			}
		}
	}
}

sub niceDate
{
	my $esecs = shift;

	my ($sec,$min,$hour,$day,$mon,$year,$wday,$yday,$isdst) = localtime($esecs);
	$mon++;
	$year += 1900;

	my $ts = sprintf("%4d.%02d.%02d/%02d:%02d", $year, $mon, $day, $hour, $sec);

	$ts;
}

sub dumpClassdata
{
	my ($cd, $str) = @_;

	my $lev = 0;
	print "### DUMP ($str) CLASS_DATA=[$cd] ###\n";
	dumpHier($cd, 1);
}

sub dumpHier
{
	my ($thing, $lev) = @_;
	my $indent = "\t" x $lev;
	#print "dumpHier: lev = $lev, indent=[$indent]\n";

	if (ref($thing)) {
		if (ref($thing) eq 'HASH') {
			foreach my $k (keys(%{$thing})) {
				if (ref($thing->{$k})) {
					print "${indent}$k = (hash)\n";
					dumpHier($thing->{$k}, $lev + 1);
				} else {
					print "${indent}$k = [$thing->{$k}]\n";
				}
			}
		} elsif (ref($thing) eq 'ARRAY') {
			foreach my $k (@{$thing}) {
				if (ref($thing->[$k])) {
					print "${indent}$k = (array)\n";
					dumpHier($thing->[$k], $lev + 1);
				} else {
					print "${indent}$k = [$thing->[$k]]\n";
				}
			}
		} else {
			die "ref=[".ref($thing)."]. huh?";
		}
	} else {
		print "${indent}$thing\n";
	}
	#print "(leaving dumpHier, lev=$lev)\n";
}

sub dumpFormat
{
	my ($cd, $field, $value) = @_;

	defined($value) || return 'NULL';

	my $type = $cd->{priv_keys}->{$field}->{type};
	if ($type eq 'time') { return niceDate($value); }

	if ($value =~ /\n/) {
		return join("\n".' 'x30, split(/\n/, $value));
	}

	return $value;
}

sub formatDiary
{
	my $diary_ref = shift;

	my $out;
	foreach my $lent (@{$diary_ref}) {
		my ($u, $t, $v) = @{$lent}{'user', 'timestamp', 'value'};
		$t = localtime($t);
		$out .= "## $u, $t\n$v\n";
	}

	$out;
}

#------------------------------------------------------------------
# schemaChanged() compares the $CLASS_DATA hierarchy to the ARS
# schema data.  If there is a discrepency a description thereof
# will be returned.  If there is no discrepency, an empty scalar
# is returned.
#
# schemaChanged() can be called periodically to check whether
# the ARS schema has been modified without fixing the corresponding
# subclass.
#------------------------------------------------------------------
sub schemaChanged
{
	my $cd = shift;

	my $debug = 0;

	my $schema = $cd->{schema};

	my %fields = ars_GetFieldTable($CTL, $schema);
	
	if (defined($fields{undef()})) { die $ars_errstr; }

	# get list of all $fids I know about
	my %my_fids;
	foreach my $tmp_fid (keys(%{$cd->{reverse_map}})) {
		$my_fids{$tmp_fid} = 1;
	}

	my $errors;
	foreach my $field (keys(%fields)) {
		$debug && print "checking field [$field]\n";
		my $fid = $fields{$field};
		if (!exists($cd->{reverse_map}->{$fid})) { next; }
		$debug && print "  field exists in private data\n";
		delete($my_fids{$fid});
		my $pkey = $cd->{reverse_map}->{$fid};
		$debug && print "  private key = [$pkey]\n";
		my $lab = $cd->{priv_keys}->{$pkey}->{real};

		$debug && print "  checking if labels match\n";
		if ($field ne $lab) {
			$errors .= "[$field]: label changed from [$lab] to [$field]\n";
		}
		my $finfo = ars_GetField($CTL, $schema, $fid);
		if (!defined($finfo)) { die "$ars_errstr"; }

		$debug && print "  checking if types match\n";
		if ($finfo->{dataType} ne $cd->{priv_keys}->{$pkey}->{type}) {
			$errors .= "[$field]: type changed from [$cd->{priv_keys}->{$pkey}->{type}] to [$finfo->{dataType}]\n";
			next;
		}
		if ($finfo->{dataType} ne 'enum') { next; }

		$debug && print "  checking if enum counts match\n";
		if (@{$finfo->{limit}} != keys(%{$cd->{priv_keys}->{$pkey}->{enum}})) {
			my $old_num = keys(%{$cd->{priv_keys}->{$pkey}->{enum}});
			my $new_num = @{$finfo->{limit}};
			$errors .= "[$field]: number of enums changed from [$old_num] to [$new_num]\n";
			next;
		}

		$debug && print "  checking if enum labels match\n";
		my $ind = 0;
		my $end = $#{$finfo->{limit}};
		$debug && print "end = [$end]\n";
		while($ind <= $#{$finfo->{limit}}) {
			$debug && print "ind = [$ind]\n";
			my $ars_lab = $finfo->{limit}->[$ind];
			my $p_ekey = $cd->{priv_keys}->{$pkey}->{enum_rev}->{$ind};
			my $saved_lab = $cd->{priv_keys}->{$pkey}->{enum}->{$p_ekey}->[1];
			$debug && print "comparing [$ars_lab] and [$saved_lab]\n";
			if ($ars_lab ne $saved_lab) {
				$errors .= "[$field]: enum label [$ind] changed from [$saved_lab] to [$ars_lab]\n";
				last;
			}
			$ind++;
		}
	}

	if (%my_fids) {
		$errors .= "fields have disappeared from ARS:";
		#join(', ', (keys(%my_fids)));
		foreach my $fid (keys(%my_fids)) {
			my $pkey = $cd->{reverse_map}->{$fid};
			my $real = $cd->{priv_keys}->{$pkey}->{real};
			$errors .= " '$real' (id=$fid);";
		}
		$errors .= "\n";
	}

	return $errors;
}

# this is ONLY for ars-gen-subclass
sub _arsControl
{
	return $CTL;
}

sub stackTrace
{
	my ($pack, $file, $line);
	my $lev = 1;
	while(($pack, $file, $line) = caller($lev)) {
		print "  $file:$line\n";
		$lev++;
	}
}

sub _stackTrace($;$) {
	my($startlev,$msg) = @_;
	my($lev) = ($startlev ? $startlev : 0);
	my(@list);
	my($out);

	my($pack,$file,$line,$sub,$eval,$require);

	($pack,$file,$line,$sub) = (caller(++$lev))[0..3];
	($pack,$file,$line,$sub) = (caller(0))[0..3];
	$sub =~ s/^[^:]*(::)*//;
	push(@list,sprintf("%-30s stackTrace initiated %s","$file:$line,",$msg));

	while (($pack,$file,$line,$sub,undef,undef,$eval,$require) =
		caller($lev++))
	{
		if ($pack eq 'main') {
			$pack = basename($0);
		}
		push(@list, sprintf("\t%-30s [from %s:%s]","$sub(),",$file,$line));
	}
	join("\n",@list)."\n";
}

sub logmsg
{
	my $msg = shift;
	print "LOG: $msg\n";
}

1;
